# coding: utf-8
"""pypel.cli, CLI interface for pypel.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import argparse
import time
import os
import warnings

from datetime import datetime

try:
    from pygments.console import ansiformat
except ImportError:
    ansiformat = None

from pypel import get_version
try:
    from .gpg import sign, verify
    gnupg = True
except ImportError:
    gnupg = False
from .models import (delete_metadata, set_metadata, make_receipt,
                     DoesNotExist, IsADirectory, ImageNotSupported)

PYPELKEY = os.environ.get('PYPELKEY')


class Row(dict):
    """Representation of a receipt as a row for terminal output."""

    FLOAT_PRECISION = '.2'

    def __init__(self, dct=None):
        if dct is not None:
            self.update(dct)

    def len(self, key=None):
        """Return len of specified field"""
        if key is not None:
            if isinstance(self[key], float):
                fmt = '%%%sf' % Row.FLOAT_PRECISION
                return len(str(fmt % self[key]))

            return len(str(self[key]))

    def format(self, key, align='', precision='', type_=''):
        if isinstance(self[key], float):
            align = '>'
            precision = Row.FLOAT_PRECISION
            type_ = 'f'

        return '{%s:%s{%s}%s%s}' % (key,
                                    align,
                                    key + '_len',
                                    precision,
                                    type_)


class Table(object):

    def __init__(self):
        self.max_len = {}
        self.rows = []

    def add_row(self, row):
        self.rows.append(row)

        if not self.max_len:
            self.max_len.update({column + '_len': 0 for column in row})

        for column in row:
            key_len = column + '_len'
            self.max_len[key_len] = max((row.len(column),
                                         self.max_len[key_len]))

    def to_string(self,
                  color=False,
                  fields_order=None,
                  sep=' -- ',
                  verify=False):

        if fields_order is None:
            fields_order = ['file', 'price', 'retailer', 'note']

        for row in self.rows:

            fmt_str = sep.join([row.format(field) for field in fields_order])

            if verify and not color:
                fmt_str += ' | {}'.format(row['verified'])

            if verify and color:
                if ansiformat:
                    if row['verified']:
                        fmt_str = ansiformat('green', fmt_str)
                    else:
                        fmt_str = ansiformat('red', fmt_str)
                else:
                    warnings.warn('You must install pygments to have colored '
                                  'output.')

            fields = row.copy()
            fields.update(self.max_len)
            yield fmt_str.format(**fields)


def make_parsers():
    """Create the parsers for the CLI tool."""

    parser = argparse.ArgumentParser(description='Easy receipts management.',
                                     version=get_version())
    subparsers = parser.add_subparsers(dest='command_name', help='commands')

    # A show command
    show_parser = subparsers.add_parser('show',
                                        help='Show receipts\' metadata')
    show_parser.add_argument('-v', '--verify', action='store_true',
                             help='verify receipts')
    show_parser.add_argument('-c', '--color', action='store_true',
                             help='colorize the output')
    show_parser.set_defaults(action=do_show)

    # A set command
    set_parser = subparsers.add_parser('set', help='Set receipts\' metadata')
    set_parser.add_argument('-p', '--price', action='store', type=float,
                            help='set receipts\' price')
    set_parser.add_argument('-r', '--retailer', action='store', type=str,
                            help='set receipts\' retailer')
    set_parser.add_argument('-n', '--note', action='store', type=str,
                            help='set receipts\' note')
    set_parser.set_defaults(action=do_set)

    # A delete command
    del_parser = subparsers.add_parser('del',
                                       help='Delete receipts\' metadata')
    del_parser.add_argument('-p', '--price', action='store_true',
                            help='delete receipts\' price')
    del_parser.add_argument('-r', '--retailer', action='store_true',
                            help='delete receipts\' retailer')
    del_parser.add_argument('-n', '--note', action='store_true',
                            help='delete receipts\' note')
    del_parser.set_defaults(action=do_del)

    # A sum command
    sum_parser = subparsers.add_parser('sum', help='Sum receipts\' price')
    sum_parser.set_defaults(action=do_sum)

    # A gpg command
    gpg_parser = subparsers.add_parser('gpg', help='Sign or verify receipts')
    gpg_group = gpg_parser.add_mutually_exclusive_group()
    gpg_group.add_argument('-s', '--sign', action='store_true',
                           help='sign receipts')
    gpg_group.add_argument('-v', '--verify', action='store_true',
                           help='verify receipts')
    gpg_parser.set_defaults(action=do_gpg)

    all_subparsers = dict(
        show_parser=show_parser,
        set_parser=set_parser,
        del_parser=del_parser,
        sum_parser=sum_parser,
        gpg_parser=gpg_parser)

    # HACK: This can be fixed when http://bugs.python.org/issue9540 will be
    # closed.
    for subparser in all_subparsers:
        all_subparsers[subparser].add_argument('receipts',
                                               metavar='receipt',
                                               nargs='+',
                                               help='one or more receipts in a'
                                                    ' supported format')

    return parser, all_subparsers


def receipts(args):
    for receipt_file in args.receipts:

        try:
            receipt = make_receipt(receipt_file)
            yield receipt
        except DoesNotExist as e:
            print('{}: {}'.format(receipt_file, e))
            continue
        except IsADirectory as e:
            print('{}: {}'.format(receipt_file, e))
            continue
        except ImageNotSupported as e:
            # Skip if receipt_file is not a supported file.
            continue


def do_show(args):

    table = Table()

    for receipt in receipts(args):
        row = Row(receipt.asdict())

        # Verify signature for the receipt if needed. If signature is
        # missing `verified' must be False.
        if args.verify:
            if not gnupg:
                warnings.warn('You must install gnupg module to sign and'
                              ' verify receipts.')
            try:
                verified = verify(receipt.file).valid
            except (ValueError, IOError, NameError):
                verified = False

            row.update(dict(verified=verified))

        table.add_row(row)

    print('\n'.join(table.to_string(color=args.color, verify=args.verify)))


def do_set(args):
    for receipt in receipts(args):
        set_metadata(receipt, args.price, args.retailer, args.note)


def do_del(args):
    for receipt in receipts(args):
        delete_metadata(receipt, args.price, args.retailer, args.note)


def do_sum(args):
    price_sum = 0
    for receipt in receipts(args):
        if receipt.price is not None:
            price_sum += receipt.price

    print('{0:.2f}'.format(price_sum))


def do_gpg(args):
    if gnupg:
        for receipt in receipts(args):
            if args.sign:
                sign(receipt.file, keyid=PYPELKEY)

            if args.verify:
                try:
                    verified = verify(receipt.file)
                    if verified:
                        print('Good signature from "{}"'.format(
                              verified.username))
                        d = datetime.fromtimestamp(float(verified.timestamp))
                        print('Signature made {} {} using key ID {}'.format(
                              d.isoformat(' '),
                              time.tzname[time.daylight],
                              verified.key_id))
                except ValueError as err:
                    print('{}: {}'.format(receipt.file, err))
                except IOError as err:
                    print('{}: {}'.format(err.filename, err.strerror))
    else:
        warnings.warn('You must install gnupg module to sign and verify'
                      ' receipts.')


def main():

    parser, subparsers = make_parsers()
    args = parser.parse_args()

    if args.command_name == 'set':
        if args.price is None and args.retailer is None and args.note is None:
            subparsers['set_parser'].error('You must provide at least '
                                           '--price or --retailer or --note')
    elif args.command_name == 'gpg':
        if not args.sign and not args.verify:
            subparsers['gpg_parser'].error('You must provide at least '
                                           '--sign or --verify')
    args.action(args)



# coding: utf-8
"""pypel package, simple receipts management tool.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

__version_info__ = (0, 3, 0, 'final', 0)


def get_version():
    """Returns project version in a human readable form."""
    version_numbers = 2 if __version_info__[2] == 0 else 3
    version = '.'.join(str(x) for x in __version_info__[:version_numbers])

    assert __version_info__[3] in ('alpha', 'beta', 'candidate', 'final')

    release = ''

    if __version_info__[3] != 'final':
        release = __version_info__[3][0] + str(__version_info__[4])

    return version + release

__version__ = get_version()



# coding: utf-8
"""pypel.gpg, gpg wrappers for pypel.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import gnupg

gpg = gnupg.GPG()


def _sign_file_path(file_path):
    return file_path + '.sig'


def sign(file_path, keyid=None):

    with open(file_path, 'rb') as f:
        signed = gpg.sign_file(f, keyid, detach=True)

    sign_file_path = _sign_file_path(file_path)

    with open(sign_file_path, 'wb') as sign_file:
        sign_file.write(signed.data)


def verify(file_path):

    with open(_sign_file_path(file_path), 'rb') as f:
        verified = gpg.verify_file(f, file_path)

        if not verified:
            raise ValueError('Signature could not be verified!')

        return verified



# coding: utf-8
"""pypel.models, models for pypel.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import os.path

import six

from gi.repository import GExiv2

XMP_KEY_PREFIX = 'Xmp.pypel.'

SUPPORTED_EXT = ('.jpg', '.jpeg', '.png', '.eps')


class DoesNotExist(IOError):
    """The file or directory does not exist"""


class IsADirectory(IOError):
    """This is a directory so you can't use it as a receipt"""


class ImageNotSupported(IOError):
    """Image is not supported"""


class Field(object):
    """Base class for all field types"""

    casting_function = lambda self, value: value
    creation_counter = 1

    def __init__(self):
        self.name = None

        self.creation_counter = Field.creation_counter
        Field.creation_counter += 1

    @property
    def key(self):
        return XMP_KEY_PREFIX + self.name

    def __get__(self, instance, owner):
        return self.cast(instance.get_value(self.key))

    def __set__(self, instance, value):
        instance.set_value(self.key, value)

    def __delete__(self, instance):
        instance.delete_value(self.key)

    def cast(self, value):
        if value is not None:
            return self.casting_function(value)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)


class CharField(Field):
    """A field for character strings"""


class FloatField(Field):
    """A field for float numbers"""

    casting_function = float


class ModelBase(type):
    """Metaclass for models"""

    def __init__(cls, name, bases, attrs):

        if not hasattr(cls, '_fields'):
            cls._fields = []

        for name in attrs:
            if isinstance(attrs[name], Field):
                # Set field name because it is required for key generation
                setattr(attrs[name], 'name', name.title())

                cls._fields.append(attrs[name])

        cls._fields.sort(key=lambda field: field.creation_counter)

        super(ModelBase, cls).__init__(name, bases, attrs)


class Model(six.with_metaclass(ModelBase)):

    def __init__(self, file):
        self.file = file
        self._metadata = GExiv2.Metadata(file)

        if not self._metadata.has_xmp():
            self._metadata.register_xmp_namespace('pypelNamespace/', 'pypel')

    def get_value(self, key):
        try:
            return self._metadata[key]
        except KeyError:
            return None

    def set_value(self, key, value):
        self._metadata[key] = str(value)
        self._metadata.save_file()

    def delete_value(self, key):
        try:
            del self._metadata[key]
        except KeyError:
            pass
        self._metadata.save_file()

    def asdict(self):
        d = {'file': self.file}
        for field in self._fields:
            name = field.name.lower()
            d[name] = getattr(self, name)

        return d

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.file)


class Receipt(Model):
    price = FloatField()
    retailer = CharField()
    note = CharField()


def delete_metadata(receipt, price=None, retailer=None, note=None):
    """Delete XMP metadata."""

    if not price and not retailer and not note:
        del receipt.price
        del receipt.retailer
        del receipt.note

    if price:
        del receipt.price

    if retailer:
        del receipt.retailer

    if note:
        del receipt.note


def set_metadata(receipt, price=None, retailer=None, note=None):
    """Set XMP metadata."""

    if price:
        receipt.price = price

    if retailer:
        receipt.retailer = retailer

    if note:
        receipt.note = note


def make_receipt(file):
    """Check for errors and create a receipt"""

    if not os.path.exists(file):
        raise DoesNotExist('No such file or directory')
    elif os.path.isdir(file):
        raise IsADirectory('Is a directory')
    elif os.path.splitext(file)[1].lower() not in SUPPORTED_EXT:
        raise ImageNotSupported

    return Receipt(file)



# coding: utf-8
""" Tests for pypel.cli.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import unittest

from pypel.cli import Row, Table


class RowTestCase(unittest.TestCase):

    def test_empty(self):
        row = Row()

        self.assertEqual(row.len(), None)
        with self.assertRaises(TypeError):
            row.format()
        with self.assertRaises(KeyError):
            row.format('price')

    def test_float(self):
        row = Row({'price': 2.71})

        self.assertEqual(row.len('price'), 4)
        self.assertEqual(row.format('price'), '{price:>{price_len}.2f}')

    def test_int(self):
        row = Row({'price': 2})

        self.assertEqual(row.len('price'), 1)
        self.assertEqual(row.format('price'), '{price:{price_len}}')

    def test_str(self):
        row = Row({'note': 'A simply note.'})
        self.assertEqual(row.len('note'), 14)
        self.assertEqual(row.format('note'), '{note:{note_len}}')


class TableTestCase(unittest.TestCase):

    def setUp(self):
        self.table = Table()

    def test_empty(self):
        self.assertEqual(self.table.rows, [])
        self.assertEqual(self.table.max_len, {})

    def test_add_new_row(self):
        row = Row({'file': 'receipt.jpg',
                   'price': 2.71,
                   'retailer': 'カオナシ',
                   'note': 'An useless note.'})
        self.table.add_row(row)

        self.assertEqual(len(self.table.rows), 1)
        self.assertEqual(''.join(self.table.to_string()),
                         'receipt.jpg -- 2.71 -- カオナシ -- An useless note.')


class TableToStringTestCase(unittest.TestCase):

    def setUp(self):
        self.table = Table()
        row = Row({'file': 'receipt.jpg',
                   'price': 2.71,
                   'retailer': 'カオナシ',
                   'note': 'An useless note.'})
        self.table.add_row(row)

    def test_fields_order(self):
        order = ['price', 'note', 'file', 'retailer']
        self.assertEqual(''.join(self.table.to_string(fields_order=order)),
                         '2.71 -- An useless note. -- receipt.jpg -- カオナシ')

    def test_separator(self):
        sep = ' * '
        self.assertEqual(''.join(self.table.to_string(sep=sep)),
                         'receipt.jpg * 2.71 * カオナシ * An useless note.')

    def test_verify(self):
        self.table.rows[0].update(dict(verified=False))
        self.assertEqual(''.join(self.table.to_string(verify=True)),
                         'receipt.jpg -- 2.71 -- カオナシ -- An useless note.'
                         ' | False')
        self.table.rows[0].update(dict(verified=True))
        self.assertEqual(''.join(self.table.to_string(verify=True)),
                         'receipt.jpg -- 2.71 -- カオナシ -- An useless note.'
                         ' | True')



# coding: utf-8
"""pypel setup file.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import distutils.core
import os.path


from pypel import __version__


def read(filename):
    """Small tool function to read file content."""
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


classifiers = '''
Development Status :: 3 - Alpha
Environment :: Console
Intended Audience :: End Users/Desktop
License :: OSI Approved :: BSD License
Operating System :: POSIX
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Topic :: Office/Business :: Financial
'''.strip().splitlines()

distutils.core.setup(
    name = 'pypel',
    version = __version__,
    license = 'BSD',
    description = 'simple tool to manage receipts',
    long_description = read('README.rst'),
    classifiers = classifiers,
    url = 'https://mornie.org/projects/pypel/',
    author = 'Daniele Tricoli',
    author_email = 'eriol@mornie.org',
    packages = ['pypel'],
    package_dir = dict(pypel='pypel'),
    scripts = ['bin/pypel']
)



# coding: utf-8
""" Tests for pypel.models.

THIS SOFTWARE IS UNDER BSD LICENSE.
Copyright (c) 2012-2015 Daniele Tricoli <eriol@mornie.org>

Read LICENSE for more informations.
"""

import tempfile
import unittest

from PIL import Image

from pypel.models import Receipt, set_metadata, delete_metadata, make_receipt
from pypel.models import DoesNotExist, IsADirectory, ImageNotSupported


class ReceiptSetUpTestCase(unittest.TestCase):
    """Create a temporary receipt."""

    def setUp(self):

        self.tmp_file = tempfile.NamedTemporaryFile()

        self.size = 100, 100
        self.image = Image.new('RGB', self.size)
        self.image.save(self.tmp_file.name, format='PNG')

        self.receipt = Receipt(self.tmp_file.name)

    def tearDown(self):
        self.tmp_file.close()


class ReceiptMetadataTestCase(ReceiptSetUpTestCase):

    def test_price(self):
        """Set/get price for receipt."""
        self.receipt.price = 8.27
        self.assertEqual(self.receipt.price, 8.27)

        del self.receipt.price
        self.assertEqual(self.receipt.price, None)

    def test_retailer(self):
        """Set/get retailer for receipt."""
        self.receipt.retailer = 'Bazaar'
        self.assertEqual(self.receipt.retailer, 'Bazaar')

        del self.receipt.retailer
        self.assertEqual(self.receipt.retailer, None)

    def test_note(self):
        """Set/get note for receipt."""
        self.receipt.note = 'A book worth reading is worth buying.'
        self.assertEqual(self.receipt.note,
                         'A book worth reading is worth buying.')

        del self.receipt.note
        self.assertEqual(self.receipt.note, None)

    def test_set_metadata(self):
        """Test set_metadata command."""

        self.assertEqual(self.receipt.price, None)
        self.assertEqual(self.receipt.retailer, None)
        self.assertEqual(self.receipt.note, None)

        set_metadata(self.receipt, price=8.27)
        self.assertEqual(self.receipt.price, 8.27)
        self.assertEqual(self.receipt.retailer, None)
        self.assertEqual(self.receipt.note, None)

        set_metadata(self.receipt, retailer='Bazaar')
        self.assertEqual(self.receipt.retailer, 'Bazaar')
        self.assertEqual(self.receipt.note, None)

        set_metadata(self.receipt, note='This is a note.')
        self.assertEqual(self.receipt.note, 'This is a note.')

    def test_delete_metadata(self):
        """Test delete_metadata command."""

        set_metadata(self.receipt,
                     price=8.27,
                     retailer='Bazaar',
                     note='A note')

        self.assertEqual(self.receipt.price, 8.27)
        self.assertEqual(self.receipt.retailer, 'Bazaar')
        self.assertEqual(self.receipt.note, 'A note')

        delete_metadata(self.receipt, price=True)
        self.assertEqual(self.receipt.price, None)

        delete_metadata(self.receipt, retailer=True)
        self.assertEqual(self.receipt.retailer, None)

        delete_metadata(self.receipt, note=True)
        self.assertEqual(self.receipt.note, None)

        set_metadata(self.receipt,
                     price=8.27,
                     retailer='Bazaar',
                     note='A note')
        delete_metadata(self.receipt)

        self.assertEqual(self.receipt.price, None)
        self.assertEqual(self.receipt.retailer, None)
        self.assertEqual(self.receipt.note, None)

    def test_asdict(self):
        set_metadata(self.receipt,
                     price=8.27,
                     retailer='Bazaar',
                     note='A note')
        self.assertEqual(self.receipt.asdict(),
                         {'file': self.tmp_file.name,
                          'price': 8.27,
                          'retailer': 'Bazaar',
                          'note': 'A note'})


class MakeReceiptTestCase(unittest.TestCase):

    def setUp(self):
        self.tmp_file_png = tempfile.NamedTemporaryFile(suffix='.png')
        self.tmp_file_txt = tempfile.NamedTemporaryFile(suffix='.txt')

        self.size = 10, 10
        self.image = Image.new('RGB', self.size)
        self.image.save(self.tmp_file_png.name, format='PNG')

        self.receipt = Receipt(self.tmp_file_png.name)

    def tearDown(self):
        self.tmp_file_png.close()
        self.tmp_file_txt.close()

    def test_does_not_exist(self):
        """Raise DoesNotExist if specified file does not exists."""
        self.assertRaises(DoesNotExist,
                          make_receipt,
                          '/it/is/very/difficult/that/this/path/exists/really/')

    def test_is_a_directory(self):
        """Raise IsADirectory if a directory is specified."""
        self.assertRaises(IsADirectory, make_receipt, '/')

    def test_image_not_supported(self):
        """Raise ImageNotSupported if specified image file is not supported."""
        self.assertRaises(ImageNotSupported,
                          make_receipt,
                          self.tmp_file_txt.name)

    def test_receipt_creation(self):
        """Create a Receipt using make_receipt factory function."""
        r = make_receipt(self.tmp_file_png.name)
        self.assertTrue(isinstance(r, Receipt))
